#include <cstdio>
#include <cstring>
#include <vector>
#include <list>
#include <deque>
#include <cmath>
#include <map>
#include <set>

using namespace std;

struct SNDChannel
{
    enum chantype { undefined=0, sqr=1, rnd=2, tri=3, pcm=4 } type;
    unsigned char vol;     // 0..15 for any
    unsigned wavelen;      // 0..0x7FF for any
    unsigned char sample;  // or duty for sqr(2,4,8,12)
    unsigned age;          // Age of this row: most useful with pcm
    
    unsigned ins; // Filled later
};

class song
{
public:
    /* Parameters */
    unsigned rate;
    unsigned channelcount;
    /* Data */
    unsigned rownum;
    vector<SNDChannel> chanptr;
    vector<unsigned char> dpcmcache[256];

    song() : rate(0), channelcount(0), rownum(0), chanage(0)
    {
    }
    
private:
    /* Variables */
    unsigned chanage;
    
    void CheckRow()
    {
        unsigned oldrownum = rownum;
        rownum += chanage;
        if(chanptr.size() <= rownum*channelcount)
        {
            const unsigned increment = 1024;
            chanptr.resize(((rownum + increment) & ~(increment-1)) * channelcount);
        }
        
        for(unsigned tmp=0; tmp<chanage; ++tmp)
        {
            unsigned t2 = tmp+1;
            unsigned src  = oldrownum * channelcount;
            unsigned dest = src + t2 * channelcount;
            for(unsigned c=0; c<channelcount; ++c)
            {
                chanptr[dest+c] = chanptr[src+c];
                chanptr[dest+c].age = chanptr[src+c].age + t2;
            }
        }
        chanage=0;
    }
public:
    void AddAge(unsigned n)
    {
        chanage += n;
    }
    
    const SNDChannel &ChanPtr(unsigned row, unsigned chan) const
    {
        return chanptr[row * channelcount + chan];
    }

    SNDChannel &GetChan(unsigned c)
    {
        CheckRow();
        SNDChannel &tmp = chanptr[rownum * channelcount + c];
        tmp.age = 0;
        return tmp;
    }
    
    /* Delete silence from the beginning of the songs */
    void DeleteSilence()
    {
        unsigned rows=0;
        while(rows < rownum)
        {
            for(unsigned c=0; c<channelcount; ++c)
            {
                const SNDChannel &ch = ChanPtr(rows, c);
                if(ch.vol || ch.wavelen)goto done;
            }
            ++rows;
        }
    done:
        if(rows)
        {
            chanptr.erase(chanptr.begin(), chanptr.begin() + rows * channelcount);
            rownum -= rows;
            fprintf(stderr, "%u row(s) of silence deleted\n", rows);
        }
    }
};

#define CONTENTDEBUG 0
static song ReadSND(const char *fn)
{
    song song;
    FILE *fp = fopen(fn, "rb");
    
    if(!fp) { perror(fn); return song; }
    
    unsigned char Buf[64];
    fread(Buf, 16, 1, fp);
    if(Buf[0]!='S' || Buf[1]!='N' || Buf[2]!='D' || Buf[3]!=26)
    {
        fclose(fp);
        fprintf(stderr, "%s: Not a SND file\n", fn);
        return song;
    }
    
    int version = Buf[4];
    
    song.channelcount = Buf[5];
    song.rate         = Buf[6];
    
    if(!song.channelcount || song.channelcount>=0x20 || !song.rate)
    {
        fclose(fp);
        fprintf(stderr, "%s: Improper settings in file\n", fn);
        return song;
    }
    
    for(;;)
    {
        int c = fgetc(fp);
        if(c == EOF)break;
#if CONTENTDEBUG
        fprintf(stderr, "%02X  ", c);
#endif
        if(c == 0xFB)
        {
            int n = fgetc(fp), sl = fgetc(fp), sh = fgetc(fp);
            song.dpcmcache[n].resize(sl + sh*256);
            fread(&song.dpcmcache[n][0], 1, song.dpcmcache[n].size(), fp);
            continue;
        }
        if(c == 0xFE)
        {
            int c = fgetc(fp), t = fgetc(fp);
            
#if CONTENTDEBUG
            fprintf(stderr, "\n");
#endif
            SNDChannel &tmp = song.GetChan(c);
            tmp.type = (SNDChannel::chantype)t;
            continue;
        }
        if(c == 0xFF)
        {
#if CONTENTDEBUG
            fprintf(stderr, "\n");
#endif
            song.AddAge(1);
            continue;
        }
        if(c == 0xFD)
        {
#if CONTENTDEBUG
            fprintf(stderr, "\n");
#endif
            song.AddAge(fgetc(fp));
            continue;
        }
        if(version == 3 &&
           ((c & 0xF0) == 0x20
         || (c & 0xF0) == 0x40
         || (c & 0xF0) == 0x80
         || (c & 0xF0) == 0xC0))
        {
            int duty = c>>4, vol=c&15;
            int c=fgetc(fp), wl=fgetc(fp), wh=fgetc(fp);
            
            SNDChannel &tmp = song.GetChan(c);
            tmp.wavelen = wl + wh*256;
            tmp.vol = vol;
            tmp.sample = duty;
            continue;
        }
        if((unsigned)c < song.channelcount)
        {
            int wl = fgetc(fp), wh = fgetc(fp), vol = fgetc(fp);
            
#if CONTENTDEBUG
            fprintf(stderr, "\b\b\b\b[%u,%u,%u]  ",c,vol,wl+wh*256);
#endif
            SNDChannel &tmp = song.GetChan(c);
            tmp.wavelen = wl + wh*256;
            
            if(tmp.type == SNDChannel::undefined)
            {
                int t = (c==0 || c==1) ? SNDChannel::sqr : (c==2 ? SNDChannel::tri : SNDChannel::rnd);
                if(t == SNDChannel::sqr)tmp.sample = 4;
                tmp.type = (SNDChannel::chantype)t;
            }

            if(version==1) { vol /= 16; tmp.wavelen = 110000 / (wl+wh*256); }
            if(version==1 && c>1) tmp.wavelen /= 2;
            
            if(tmp.type == SNDChannel::pcm)
            {
                tmp.sample = vol;
                tmp.vol = vol ? 6 : 0;
            }
            else
                tmp.vol = vol;
            continue;
        }
    }
    song.GetChan(0);
    song.DeleteSilence();
    fclose(fp);
    fprintf(stderr, "Got %u rows (at rate %u, %u seconds)\n",
        song.rownum, song.rate, (song.rownum + song.rate - 1) / song.rate);
    return song;
}

struct sampleitem
{
    int duty; /* Duty at this row */
    /* Frequency changes to the original */
    int wavelenadd;
    long wavelenmul;
    
    /* Volume changes to the original */
    int voladd;
    long volmul;
};
struct sample
{
    SNDChannel::chantype type;
    unsigned beginrow;
    unsigned basewavelen;
    unsigned char basevol;

    // Instrumentista on tss 7 elementin
    // alkuptk ja 7 elementin loppuptk.
    deque<sampleitem> alkutree;
    /*
    deque<sampleitem> lopputree;
    */
    
    sample() : beginrow(0),basewavelen(0),basevol(0) {}
    
    enum comparetype { undefined=0, add, mul };
    /* Whether the wavelen is compared with multiplication or addition */
    comparetype wavelentype;
    /* Whether the volume is compared with multiplication or addition */
    comparetype voltype;

    /* For noise channels, duty doesn't count */
    /* For triangle channels, duty and volume don't count */
    
    const bool dutytest() const { return type == SNDChannel::sqr; }
    const bool volutest() const { return type != SNDChannel::tri; }
    const bool wavetest() const { return type != SNDChannel::sqr; }
};

static vector<sample> repository;
static double TickSpacer;
static vector< set<unsigned> > BeatList;

struct SampleMatchType
{
    unsigned length;
    bool waveaddmatch, wavemulmatch;
    bool voladdmatch, volmulmatch;
    
    SampleMatchType() : length(0),
      waveaddmatch(true), wavemulmatch(true),
      voladdmatch(true), volmulmatch(true) { }
    SampleMatchType(bool) : length(0) { }
};

static SampleMatchType TreeCompare(
  const sample &book, const deque<sampleitem> &booktree,
  const sample &,     const deque<sampleitem> &datatree)
{
    typedef deque<sampleitem>::const_iterator itemit;

    const bool dutytest = book.dutytest();
    const bool volutest = book.volutest();
    const bool wavetest = book.wavetest();
    
    itemit bookit = booktree.begin();
    itemit datait = datatree.begin();
    
    // Note: data has all "types" undefined.
    // If the book has them defined, they master the comparison.
    
    SampleMatchType match;
    
    while(bookit != booktree.end()
       && datait != datatree.end())
    {
        if(dutytest && bookit->duty != datait->duty) return false;
        if(wavetest)
        {
            const bool addmatch = bookit->wavelenadd == datait->wavelenadd;
            const bool mulmatch = bookit->wavelenmul == datait->wavelenmul;
/*
            fprintf(stderr, "[Len %u, %u%s%u %ld%s%ld]\n", match.length,
                bookit->wavelenadd,addmatch?"==":"!=", datait->wavelenadd,
                bookit->wavelenmul,mulmatch?"==":"!=", datait->wavelenmul);
*/
            if(!addmatch) { if(book.wavelentype == sample::add)return false; if(!match.wavemulmatch)return false; match.waveaddmatch = false; }
            if(!mulmatch) { if(book.wavelentype == sample::mul)return false; if(!match.waveaddmatch)return false; match.wavemulmatch = false; }
        }
        if(volutest)
        {
            const bool addmatch = bookit->voladd == datait->voladd;
            const bool mulmatch = bookit->volmul == datait->volmul;
/*
            fprintf(stderr, "[Len %u, %u%s%u %ld%s%ld]\n", match.length,
                bookit->voladd,addmatch?"==":"!=", datait->voladd,
                bookit->volmul,mulmatch?"==":"!=", datait->volmul);
*/
            if(!addmatch) { if(book.voltype == sample::add)return false; if(!match.volmulmatch)return false; match.voladdmatch = false; }
            if(!mulmatch) { if(book.voltype == sample::mul)return false; if(!match.voladdmatch)return false; match.volmulmatch = false; }
        }
        ++bookit;
        ++datait;
        ++match.length;
        #if 0
        if(match.length >= 7)
        {
            while(bookit != book.tree.end()
               && datait != data.tree.end())
            {
                ++bookit;
                ++datait;
                ++match.length;
            }
            break;
        }
        #endif
    }
    // Found a match. {wave,vol}{add,mul}match could be used
    // now to tell which kind of matching we should be doing.
    return match;
}

struct SampleMatchPairType
{
    SampleMatchType alku;
    /*SampleMatchType loppu;*/
    
    SampleMatchPairType() :  alku(false)/*, loppu(false)*/ { }
    SampleMatchPairType(const SampleMatchType &a/*,
                        const SampleMatchType &b*/) : alku(a)/*, loppu(b)*/ { }
    SampleMatchPairType(bool v) : alku(v)/*, loppu(v)*/ { }
};

// Test whether the "book" defines "data"
static SampleMatchPairType TestMatch(const sample &book, const sample &data)
{
    // Different type samples can never match
    if(book.type != data.type)
    {
        return false;
    }
    
    return SampleMatchPairType
     (TreeCompare(book, book.alkutree, data, data.alkutree)/*,
      TreeCompare(book, book.lopputree, data, data.lopputree)*/
    );
}

static void DumpSampleTree(const sample &ins, const deque<sampleitem> &tree)
{
    deque<sampleitem>::const_iterator i;
    for(i=tree.begin(); i!=tree.end(); ++i)
    {
        fprintf(stderr, "     ");
        if(ins.dutytest())
            fprintf(stderr, " Duty %d ", i->duty);
        if(ins.volutest())
        {
            fprintf(stderr, " Volume");
            if(ins.voltype!=sample::mul)fprintf(stderr, " %+d", i->voladd);
            if(ins.voltype!=sample::add)fprintf(stderr, " *%ld", i->volmul);
            fprintf(stderr, " ");
        }
        if(ins.wavetest())
        {
            fprintf(stderr, " Wavelen");
            if(ins.wavelentype!=sample::mul)fprintf(stderr, " %+d", i->wavelenadd);
            if(ins.wavelentype!=sample::add)fprintf(stderr, " *%ld", i->wavelenmul);
            fprintf(stderr, " ");
        }
        fprintf(stderr, "\n");
    }
}

static void DumpSample(const sample &ins)
{
    fprintf(stderr, "   Type %s\n",
        (ins.type==SNDChannel::sqr)?"SQR"
       :(ins.type==SNDChannel::tri)?"TRI"
       :(ins.type==SNDChannel::rnd)?"RND"
       :(ins.type==SNDChannel::pcm)?"PCM":"???");
    if(ins.volutest())
        fprintf(stderr, "   Base volume: %u\n"
                        "   Volume test type: %s\n",
            ins.basevol,
            ins.voltype==sample::undefined?"UNDEFINED"
           :ins.voltype==sample::add?"ADD"
           :ins.voltype==sample::mul?"MUL":"???");
    if(ins.wavetest())
        fprintf(stderr, "   Base wavelen: %u\n"
                        "   Wavelen test type: %s\n",
            ins.basewavelen,
            ins.wavelentype==sample::undefined?"UNDEFINED"
           :ins.wavelentype==sample::add?"ADD"
           :ins.wavelentype==sample::mul?"MUL":"???");
    //fprintf(stderr, "   Length: %u samples\n", ins.tree.size());
    
    if(ins.type == SNDChannel::sqr || ins.type == SNDChannel::rnd)
    {
        struct { int len, power; } attack, decay;
        attack.len = attack.power = 0;
        for(unsigned a=1; a<ins.alkutree.size(); ++a)
        {
            const sampleitem &s = ins.alkutree[a];
            if(s.voladd <= ins.alkutree[a-1].voladd)
                break;
            attack.power = s.voladd;
            attack.len = a;
        }
        decay.len = decay.power = 0;
        for(unsigned a=1+attack.len, decaystart=a; a<ins.alkutree.size(); ++a)
        {
            const sampleitem &s1 = ins.alkutree[decaystart];
            const sampleitem &s2 = ins.alkutree[a];
            
            if(a > decaystart)
            {
                int slope = s2.voladd - ins.alkutree[a-1].voladd;
                if(slope > 0)break; // If the volume started rising again
                
                slope *= 10;
                int oldslope = decay.power*10/ decay.len;
                
                if(slope < oldslope-2 || slope > oldslope+2)
                {
                    // Dramatic change, break
                }
            }
            
            decay.len   = a-decaystart;
            decay.power = -(s2.voladd - s1.voladd);
        }
        fprintf(stderr, "  Attack(len %d,power %d) Decay(len %d,power %d)\n",
            attack.len,attack.power,
            decay.len, decay.power);
        
        switch(ins.type)
        {
            case SNDChannel::sqr:
                if(attack.power >= 3 && decay.len > 3)
                    fprintf(stderr, "    (Blow)\n");
                else if(!attack.power && decay.power)
                    fprintf(stderr, "    (Hit)\n");
                else if(!attack.power && !decay.power)
                    fprintf(stderr, "    (Flute)\n");
                else
                    fprintf(stderr, "    (String)\n");
                break;

            case SNDChannel::rnd:
                if(attack.power >= 3 && decay.len > 3)
                    fprintf(stderr, "    (Blow)\n");
                else if(!attack.power && decay.power)
                    fprintf(stderr, "    (Snare)\n");
                else
                    fprintf(stderr, "    (Snare)\n");
                break;
            default:
                break;
        }
    }
    fprintf(stderr, "  Begins:\n");
    DumpSampleTree(ins, ins.alkutree);
    /*fprintf(stderr, "  Ends:\n");
    DumpSampleTree(ins, ins.lopputree);
    */
}

static unsigned AnalyzeSample(const sample &smp)
{
    /*
    fprintf(stderr, "--------------------\n");
    DumpSample(smp);
        */

    SampleMatchPairType bestmatch;
    unsigned bestmatchnum=0;
    for(unsigned a=0; a<repository.size(); ++a)
    {
        //fprintf(stderr, "COMPARING TO INS %u\n", a+1);
        SampleMatchPairType match = TestMatch(repository[a], smp);
        
        if(!match.alku.length/* || !match.loppu.length*/)
            continue;
        
        if(match.alku.length > bestmatch.alku.length
        /*
        || (match.alku.length == bestmatch.alku.length
         && match.loppu.length > bestmatch.loppu.length)
        */
          )
        {
            bestmatch    = match;
            bestmatchnum = a;
        }
    }
    
/*
    if(!bestmatch.length)
        fprintf(stderr, "->No match!\n");
    else
    {
        fprintf(stderr, "->Best match:\n"
                        "  Length %u\n", bestmatch.length);
        if(smp.wavetest())
            fprintf(stderr, "  Wavelen match: %s,%s\n",
                bestmatch.waveaddmatch?"Add":"",
                bestmatch.wavemulmatch?"Mul":"");
        if(smp.volutest())
            fprintf(stderr, "  Volume match: %s,%s\n",
                bestmatch.voladdmatch?"Add":"",
                bestmatch.volmulmatch?"Mul":"");
    }
*/
    
    if(bestmatch.alku.length)
    {
        sample &book = repository[bestmatchnum];
        sample newsample = book;
        bool changes = false;
        /* Upgrade the sample cache with possibly available matching type information */
        if(newsample.volutest() && newsample.voltype == sample::undefined)
        {
            if(bestmatch.alku.voladdmatch && !bestmatch.alku.volmulmatch)
            {
                changes = true;
                newsample.voltype = sample::add;
            }
            else if(bestmatch.alku.volmulmatch && !bestmatch.alku.voladdmatch)
            {
                changes = true;
                newsample.voltype = sample::mul;
            }
        }
        if(newsample.wavetest() && newsample.wavelentype == sample::undefined)
        {
            if(bestmatch.alku.waveaddmatch && !bestmatch.alku.wavemulmatch)
            {
                changes = true;
                newsample.wavelentype = sample::add;
            }
            else if(bestmatch.alku.wavemulmatch && !bestmatch.alku.waveaddmatch)
            {
                changes = true;
                newsample.wavelentype = sample::mul;
            }
        }
        
        if(smp.alkutree.size() > newsample.alkutree.size())
            newsample.alkutree = smp.alkutree;
        /*
        if(smp.lopputree.size() > newsample.lopputree.size())
            newsample.lopputree = smp.lopputree;
        */
        
        /*if(changes)
        {
            repository.push_back(newsample);
            return repository.size();
        }
        else*/
        {
            //fprintf(stderr, "UPDATED INSTRUMENT %u:\n", bestmatchnum+1);
            //DumpSample(newsample);
            book = newsample;
        }
        return bestmatchnum+1;
    }
    else
    {
        sample tmp = smp;
        repository.push_back(tmp);
        
        //fprintf(stderr, "CACHED INSTRUMENT %u:\n", repository.size());
        //DumpSample(tmp);
        
        /* None was found */
        return repository.size();
    }
}

static bool IsTickPoint(const SNDChannel *prev, const SNDChannel *now)
{
    double wlendiff;
    if(!prev)return now;
    
    if(now->wavelen == 0 || prev->wavelen == 0)
        wlendiff = 0;
    else if(now->wavelen > prev->wavelen)
        wlendiff = now->wavelen / (double) prev->wavelen;
    else
        wlendiff = prev->wavelen / (double) now->wavelen;
    
    if(!prev->vol
    || now->vol >= prev->vol+4
    || (prev->vol > 1 && now->vol >= prev->vol*3)
    || (now->type != SNDChannel::rnd && wlendiff >= 1.057))
    {
        return true;
    }
    return false;
}

static unsigned GCD2(unsigned a, unsigned b)
{
    while(a > 0) { unsigned tmp = b % a; b = a; a = tmp; }
    return b;
}
static unsigned GCD(const set<unsigned> &nums)
{
    unsigned last = 0;
    for(set<unsigned>::const_iterator i = nums.begin(); i != nums.end(); ++i)
        last = last ? GCD2(*i, last) : *i;
    return last;
}

static unsigned LCD(const set<unsigned> &nums)
{
    unsigned a=1;
    for(set<unsigned>::const_iterator i = nums.begin(); i != nums.end(); ++i)
        a *= *i;
    return a;// / GCD(nums);
}

static void AnalyzeSND(song &song)
{
    /* For DPCM channels the DPCM equipment only
     * takes care of instruments and they are rarely
     * tweaked, so we can take them just as raw */
     
    /* Often, it seems, samples start on one channel just a tick
     * before/after other channels. In these cases, it's considerable
     * to include that void tick in the sample.
     * Another way could be to move the sample backwards/forwards
     * for one tick measure...
     */

    vector<unsigned> lasttick(song.channelcount);
    for(unsigned c=0; c<song.channelcount; ++c)
        lasttick[c] = 0;
    vector<sample> cursample(song.channelcount);

    vector< vector<bool> > ticklists(song.channelcount);
    for(unsigned a=0; a<song.channelcount; ++a)
        ticklists[a].resize(song.rownum, false);
    
    BeatList.clear();
    BeatList.resize(song.channelcount);
    
    for(unsigned a=0; a<song.rownum; ++a)
    {
        for(unsigned c=0; c<song.channelcount; ++c)
        {
            vector<bool> &ticks = ticklists[c];
            const SNDChannel &tmp = song.ChanPtr(a, c);
            if(tmp.vol)
            {
                const SNDChannel *prev = a > 0 ? &song.ChanPtr(a-1, c) : NULL;
                
                sample &chansample = cursample[c];
                
                if(IsTickPoint(prev, &tmp))
                {
                    if(a != lasttick[c] + 1 || a == 1)
                    {
                        ticks[a] = true;
                        BeatList[c].insert(a);

                        if(tmp.type == SNDChannel::pcm)
                        {
                            continue;
                        }
            
                        if(chansample.alkutree.size())
                        {
                            unsigned ins = AnalyzeSample(chansample);
                            song.chanptr[chansample.beginrow * song.channelcount + c].ins = ins;
                        }
                        /* Start of a new sample! */
                        chansample.type        = tmp.type;
                        chansample.beginrow    = a;
                        chansample.basewavelen = tmp.wavelen;
                        chansample.basevol     = tmp.vol;
                        chansample.alkutree.clear();
                        /*
                        chansample.lopputree.clear();
                        */
                    }
                    lasttick[c] = a;
                }

                if(tmp.type == SNDChannel::pcm)
                {
                    continue;
                }
            
                if(chansample.basewavelen)
                {
                    sampleitem item;
                    item.duty       = tmp.sample;
                    item.wavelenadd = tmp.wavelen - chansample.basewavelen;
                    item.wavelenmul = tmp.wavelen * 2 / chansample.basewavelen;
                    item.voladd     = tmp.vol - chansample.basevol;
                    item.volmul     = tmp.vol * 2 / chansample.basevol;
                    
                    if(chansample.alkutree.size() < 7)
                    {
                        chansample.alkutree.push_back(item);
                    }
                    /*
                    chansample.lopputree.push_front(item);
                    if(chansample.lopputree.size() > chansample.alkutree.size())
                    {
                        chansample.lopputree.erase
                         (chansample.lopputree.begin() + (chansample.lopputree.size()-1));
                    }
                    */
                }
            }
        }
    }
    
    map<unsigned, unsigned> lencounts;
    
    for(unsigned b=0; b<ticklists.size(); ++b)
    {
        vector<bool> &ticks = ticklists[b];
        unsigned a, firsttick;
        for(a=0; a<ticks.size() && !ticks[a]; ++a);
        firsttick=a;
        while(++a<ticks.size())
        {
            if(ticks[a]) { ++lencounts[a-firsttick]; firsttick=a; }
        }
    }
    unsigned mostticks = 0;
    for(map<unsigned,unsigned>::const_iterator
        i=lencounts.begin(); i!=lencounts.end(); ++i)
        if(i->second > mostticks)
            mostticks = i->second;

    for(map<unsigned,unsigned>::const_iterator
        i=lencounts.begin(); i!=lencounts.end(); ++i)
        fprintf(stderr, "%u:%u ", i->first, i->second);
    fprintf(stderr, "\n");

    set<double> lens;
    set<unsigned> divisors;

    for(map<unsigned,unsigned>::const_iterator
        i=lencounts.begin(); i!=lencounts.end(); ++i)
    {
        unsigned len = i->first;
        unsigned count=i->second;
        
        if(count < mostticks/20) continue;
        
        map<unsigned,unsigned>::const_iterator j = i; ++j;
        if(j != lencounts.end() && j->first == len+1)
        {
            unsigned len2 = j->first;
            unsigned count2 = j->second;
            
            if(count2 >= mostticks/20)
            {
                unsigned maara = count + count2;
                
                double suhde = (count > count2) ? count / (double)count2 : count2 / (double)count;
                if(suhde < 30)
                {
                    double uusi = (len * count + len2 * count2) / (double)maara;
                    i = j;
                    lens.insert(uusi);
                    divisors.insert(maara);
                    continue;
                }
            }
        }

        lens.insert(len);
    }

    for(set<double>::const_iterator
        i=lens.begin(); i!=lens.end(); ++i)
        fprintf(stderr, "%g ", *i);
    fprintf(stderr, "\n");

    unsigned divisor = LCD(divisors);
    set<unsigned> intlens;
    for(set<double>::const_iterator i = lens.begin(); i != lens.end(); ++i)
        intlens.insert((unsigned) (*i * divisor));
    double step = GCD(intlens) / (double)divisor;
    
    if(step < 1.0 && ((1.0 / step) - (unsigned)(1.0 / step)) < 0.2)
        step = 1.0;
    fprintf(stderr, "Step: %g\n", step);
    
    if(step > 4) step /= 2;
    TickSpacer = step;
}

static bool IsBeat(unsigned row, unsigned chan)
{
    return BeatList[chan].find(row) != BeatList[chan].end();
}

static void DumpSND(const song &song)
{
    for(unsigned a=0; a<repository.size(); ++a)
    {
        fprintf(stderr, "Instrument %u:\n", a+1);
        DumpSample(repository[a]);
    }

    for(unsigned a=0; a<song.rownum; ++a)
    {
        for(unsigned c=0; c<song.channelcount; ++c)
        {
            const SNDChannel &tmp = song.ChanPtr(a, c);
            char Buf[64];
            int colormode=0;
            
            sprintf(Buf, "v%Xs%dw%ua%u", tmp.vol, tmp.sample, tmp.wavelen, tmp.age);
            if(tmp.ins)
                sprintf(strchr(Buf,0), "(%u)", tmp.ins);
            
            sprintf(Buf+32, "%18s", Buf);
            if(tmp.vol)
            {
                if(IsBeat(a, c))
                {
                    fputs("\33[1;34m", stderr);
                    colormode=2;
                }
                else
                    colormode=1;
            }
            else
            {
                colormode=0;
                fputs("\33[1;30m", stderr);
            }
            for(char *s=Buf+32; *s; ++s)
            {
                /*
                if(colormode==1)
                  if(*s=='v'||*s=='s'||*s=='w'||*s=='a')fputs("\33[1;37m", stderr);
                */
                fputc(*s, stderr    );
                /*
                if(colormode==1)
                  if(*s=='v'||*s=='s'||*s=='w'||*s=='a')fputs("\33[m", stderr);
                */
            }
            fputs("\33[m", stderr);
        }
        fprintf(stderr, "\n");
    }
}

static unsigned NesVoltoS3Mvol(unsigned v)
{
    static unsigned char voltable[16] =
    {0,32,38,40,
     42,46,48,50,
     52,54,56,58,
     60,62,63,64};
    return voltable[v];
}

static void Patternize(const song &song)
{
    /* TODO: Handle the sample repository */
    
    /* Handle each channel separately */
    for(unsigned chan=0; chan<song.channelcount; ++chan)
    {
        unsigned nuotti=0, basewave=0;
        
        unsigned t=0;
        unsigned lastrow=0;
        bool wason = false;
        for(double pos=0; ; pos += TickSpacer)
        {
            unsigned row = (unsigned)pos;
            if(row >= song.rownum)break;
        
            const SNDChannel &tmp = song.ChanPtr(row, chan);
            
            // vol=tmp.vol     (volume(0..15))
            // smp=tmp.sample  (sqr:duty; pcm:pcm sample no)
            // wav=tmp.wavelen (wave length)
            // age=tmp.age     (age since last touched)
            // ins=tmp.ins     (index to repository)
            // beat=beat
            const SNDChannel *beatrow = NULL;
            bool beat = false;
            while(lastrow <= row)
                if(IsBeat(lastrow++, chan))
                {
                	beatrow = &song.ChanPtr(lastrow-1, chan);
                    beat = true;
                }
            
            fprintf(stderr, "%u ", t++%4);
            
            if(beat)
            {
                static const char nuotit[12*3+1] = "C-\0C#\0D-\0D#\0E-\0F-\0F#\0G-\0G#\0A-\0A#\0B-\0";
                basewave = tmp.wavelen;
                nuotti = (unsigned)(12 * log(4000.0 / basewave) / log(2.0));
                
                char Buf[64];
                sprintf(Buf, "%s%d", nuotit+3*(nuotti%12), nuotti/12);
                fprintf(stderr, "%s", Buf);
                wason = true;
            }
            else if(wason && !tmp.vol)
            {
                fprintf(stderr, "^^^");
                wason = false;
            }
            else
                fprintf(stderr, "...");
            
            if(tmp.type == SNDChannel::pcm && beat && tmp.sample)
                fprintf(stderr, " %02d", tmp.sample + repository.size());
            else if(tmp.type == SNDChannel::pcm && beat)
                fprintf(stderr, " %02d", beatrow->sample + repository.size());
            else if(tmp.ins)
                fprintf(stderr, " %02d", tmp.ins);
            else if(beatrow && beatrow->ins)
                fprintf(stderr, " %02d", beatrow->ins);
            else
                fprintf(stderr, " ..");
            
            if(beat)
            {
                int volavg = 0, volcount=0;
                for(unsigned tmprow = row; tmprow < row+8 && tmprow < song.rownum; ++tmprow)
                {
                    if(tmprow > row && IsBeat(tmprow, chan))break;
                    const SNDChannel &tmp = song.ChanPtr(tmprow, chan);
                    if(!tmp.vol)break;
                    volavg += tmp.vol; ++volcount;
                }
                volavg /= volcount;
                fprintf(stderr, " %02d", NesVoltoS3Mvol(volavg));
            }
            else
            {
                fprintf(stderr, " %02d", NesVoltoS3Mvol(tmp.vol));
            }
            
            fprintf(stderr, "\n");
        }
        fprintf(stderr, "----------\n");
    }
}

#include "sndplayer.cc"

int main(int argc, const char *const *argv)
{
    if(argc != 2 || !strcmp(argv[1], "--help") || !strcmp(argv[1], "/?"))
    {
        puts("Simple SNDv3 player - Copyright (C) 1992,2002 Bisqwit (http://bisqwit.iki.fi/)\n"
             "Usage:\n"
             "   snd <filename.snd>\n"
            );
        return 1;
    }
    
    song song = ReadSND(argv[1]);
    
    /* TODO: Handle the snd (dump it somewhere or something) */
    
    AnalyzeSND(song);
    DumpSND(song);
    Patternize(song);
    PlaySND(song);
    
    return 0;
}
